#include <blackfin.h>
#include <builtins.h>
#include <sys/exception.h>
#include <stdint.h>
#include <string.h>
#include "bfin_uart.h"

static uint8_t sf_uart_read_buf[BFIN_UART_READ_BUF_SIZE];
static uint8_t sf_uart_read_cnt;

EX_INTERRUPT_HANDLER(s_bfin_uart_isr);
static void s_bfin_uart_isr(void);

void e_bfin_uart_init(uint32_t baudrate, uint32_t sysclk)
{
	memset(&sf_uart_read_buf, 0, sizeof(sf_uart_read_buf));
	sf_uart_read_cnt = 0;

	/* PF11/PF12sUARTɊU */
    *pPORTF_MUX &= ~(PF11 | PF12);	/* PF11,PF120ɂ */
    *pPORTF_FER |= PF11 | PF12;		/* PF11,PF12UART0ɊU */

	/* UARTCl[u */
	*pUART0_GCTL = 1;

	/* UART̐ݒ */
	*pUART0_LCR |= DLAB;
	*pUART0_DLL = (sysclk / baudrate / 16) & 0xFF;
	*pUART0_DLH = (sysclk / baudrate / 16) >> 8;
	*pUART0_LCR &= ~DLAB;

	/* UART[hݒ peB 8bit 1Xgbvrbg */
	*pUART0_LCR = WLS(8);

	/* 荞݃nhIVG10ɓo^ */
    register_handler(ik_ivg10, s_bfin_uart_isr);
	/* UARTM݂ */
	*pUART0_IER = ERBFI;
	*pSIC_IMASK |= IRQ_DMA7;

	/* DMA8(UART TX)̐ݒ */
	*pDMA8_PERIPHERAL_MAP = 0x8000;
}

int32_t e_bfin_uart_write(const uint8_t* p_write_data, uint16_t write_cnt)
{
	if(*pDMA8_CONFIG & DMAEN)
	{
		if(((*pDMA8_IRQ_STATUS & DMA_DONE) == 0u) || ((*pUART0_LSR & TEMT) == 0u)) return BFIN_UART_ERR_DMABUSY;
		*pDMA8_CONFIG = 0u;
	}
	if(write_cnt == 0) return BFIN_UART_ERR_SUCCESS;

	*pDMA8_START_ADDR = (void*)p_write_data;
	*pDMA8_X_COUNT = write_cnt;
	*pDMA8_X_MODIFY = sizeof(uint8_t);
	*pDMA8_IRQ_STATUS = DMA_DONE | DMA_ERR;
	*pDMA8_CONFIG = FLOW_STOP | DI_EN | WDSIZE_8 | DMAEN;
	*pUART0_IER |= ETBEI;
	return BFIN_UART_ERR_SUCCESS;
}

int32_t e_bfin_uart_write_chk(void)
{
	if(*pDMA8_CONFIG & DMAEN)
	{
		if(((*pDMA8_IRQ_STATUS & DMA_DONE) == 0u) || ((*pUART0_LSR & TEMT) == 0u)) return BFIN_UART_ERR_DMABUSY;
		*pDMA8_CONFIG = 0u;
		*pDMA8_IRQ_STATUS = DMA_DONE | DMA_ERR;
	}
	return BFIN_UART_ERR_SUCCESS;
}

int32_t e_bfin_uart_read(uint8_t* p_read_data, uint16_t read_cnt)
{
	int32_t l_read_cnt;
	uint32_t l_imask = cli();
	asm("%0 = MIN(%1, %2);" : "=d"(l_read_cnt) : "d"((int32_t)sf_uart_read_cnt), "d"((int32_t)read_cnt));
	memcpy(p_read_data, sf_uart_read_buf, l_read_cnt);
	sf_uart_read_cnt -= (uint8_t)l_read_cnt;
	memmove(&sf_uart_read_buf[0], &sf_uart_read_buf[l_read_cnt], sf_uart_read_cnt);
	sti(l_imask);
	return l_read_cnt;
}

int32_t e_bfin_uart_read_buf_cnt(void)
{
	uint32_t l_imask;
	int32_t  l_remain;
	l_imask = cli();
	l_remain = (int32_t)sf_uart_read_cnt;
	sti(l_imask);
	return l_remain;
}

int32_t e_bfin_uart_print(const char *p_buf)
{
	return e_bfin_uart_write((const uint8_t*)p_buf, strlen(p_buf));
}

EX_INTERRUPT_HANDLER(s_bfin_uart_isr)
{
	uint16_t l_lsr = *pUART0_LSR;
	
	if(l_lsr & DR)
	{
		if(sf_uart_read_cnt == BFIN_UART_READ_BUF_SIZE)
		{
			memmove(&sf_uart_read_buf[0], &sf_uart_read_buf[1], BFIN_UART_READ_BUF_SIZE - 1);
			sf_uart_read_buf[BFIN_UART_READ_BUF_SIZE - 1] = *pUART0_RBR;
		}
		else
		{
			sf_uart_read_buf[sf_uart_read_cnt] = *pUART0_RBR;
			sf_uart_read_cnt++;
		}
	}
}
